#include <cmath>
#include <sigc++/bind.h>
#include "display.h"
#include "display_private.h"
#include "window.h"

using namespace std;

namespace Msp {
namespace Graphics {

void Display::add_window(Window &wnd)
{
	priv->windows[wnd.get_private().window] = &wnd;
	wnd.signal_got_focus.connect(sigc::bind(sigc::mem_fun(this, &Display::window_got_focus), sigc::ref(wnd)));
	wnd.signal_lost_focus.connect(sigc::mem_fun(this, &Display::window_lost_focus));
}

void Display::remove_window(Window &wnd)
{
	priv->windows.erase(wnd.get_private().window);
	if(&wnd==focus_window)
		focus_window = 0;
}

const VideoMode &Display::get_desktop_mode() const
{
	if(!primary_monitor || !primary_monitor->desktop_settings.mode)
		throw logic_error("no desktop mode");
	return *primary_monitor->desktop_settings.mode;
}

void Display::restore_mode()
{
	for(list<Monitor>::const_iterator i=monitors.begin(); i!=monitors.end(); ++i)
		if(i->desktop_settings.mode)
			set_mode(*i->desktop_settings.mode, false);
}

const VideoMode *Display::find_mode(const VideoMode &mode, float rate_tolerance) const
{
	const VideoMode *closest = 0;
	float rate_diff = 0.0f;

	float target_rate = mode.rate;
	if(!target_rate && mode.monitor)
		if(const VideoMode *current_mode = mode.monitor->current_settings.mode)
			target_rate = current_mode->rate;

	for(list<VideoMode>::const_iterator i=modes.begin(); i!=modes.end(); ++i)
	{
		if(mode.monitor && i->monitor!=mode.monitor)
			continue;
		// Only check rate tolerance against an explicitly requested rate
		if(mode.rate && (i->rate<mode.rate-rate_tolerance || i->rate>mode.rate+rate_tolerance))
			continue;
		if(i->width!=mode.width || i->height!=mode.height)
			continue;

		float ref_rate = target_rate;
		if(!ref_rate)
			if(const VideoMode *current_mode = i->monitor->current_settings.mode)
				ref_rate = current_mode->rate;

		float d = abs(i->rate-ref_rate);
		if(!closest || d<rate_diff)
		{
			closest = &*i;
			rate_diff = d;
		}
	}

	return closest;
}

const VideoMode *Display::find_mode(unsigned width, unsigned height) const
{
	return find_mode(VideoMode(width, height));
}

void Display::window_got_focus(Window &w)
{
	focus_window = &w;
}

void Display::window_lost_focus()
{
	focus_window = 0;
}

void Display::tick()
{
	check_error();

	Window *old_focus = focus_window;

	while(process_events()) ;

	if(old_focus && !focus_window)
		signal_lost_focus.emit();
	else if(!old_focus && focus_window)
		signal_got_focus.emit();
}

} // namespace Graphics
} // namespace Msp
